# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Literal

ChildrenType = Literal["inclusive", "exclusive"]

privacy_budget_count: int = 0

class ProvenanceEntity:
    parents                 : list[ProvenanceEntity]
    children_type           : ChildrenType
    root_name               : str
    consumed_privacy_budget : float
    depth                   : int

    def __init__(self, parents: list[ProvenanceEntity], children_type: ChildrenType, root_name: str | None = None):
        assert len(parents) > 0 or root_name is not None
        self.parents                 = parents
        self.children_type           = children_type
        self.root_name               = root_name if root_name is not None else parents[0].root_name
        self.consumed_privacy_budget = 0
        self.depth                   = max([p.depth for p in parents]) + 1 if len(parents) > 0 else 0

    def consume_privacy_budget(self, eps: float) -> None:
        assert eps >= 0
        self.consumed_privacy_budget += eps

provenance_roots : dict[str, ProvenanceEntity] = {}

def new_provenance_root(name: str) -> ProvenanceEntity:
    global provenance_roots

    if name in provenance_roots:
        return provenance_roots[name]
    else:
        pe = ProvenanceEntity([], "inclusive", root_name=name)
        provenance_roots[name] = pe
        return pe

def new_provenance_node(parents: list[ProvenanceEntity], children_type: ChildrenType) -> ProvenanceEntity:
    assert len(parents) > 0
    return ProvenanceEntity(parents, children_type)

def get_provenance_root(name: str) -> ProvenanceEntity:
    global provenance_roots

    if name not in provenance_roots:
        raise ValueError(f"Name '{name}' does not exist")

    return provenance_roots[name]

def are_exclusive_siblings(pes: list[ProvenanceEntity]) -> bool:
    assert len(pes) > 0
    return all([len(pe.parents) == 1 and \
                pes[0].parents[0] == pe.parents[0] and \
                pe.parents[0].children_type == "exclusive" for pe in pes])

def consume_privacy_budget(pes: list[ProvenanceEntity], eps: float) -> None:
    assert len(pes) > 0

    if are_exclusive_siblings(pes):
        for pe in pes:
            pe.consume_privacy_budget(eps)

        pe0 = pes[0].parents[0]
        new_cpd = max(pe0.consumed_privacy_budget, *[pe.consumed_privacy_budget for pe in pes])
        new_eps = new_cpd - pe0.consumed_privacy_budget

        if new_eps > 0:
            pe0.consume_privacy_budget(new_eps)
            consume_privacy_budget(pe0.parents, new_eps)

    elif len(pes) == 1 and len(pes[0].parents) == 0:
        assert get_provenance_root(pes[0].root_name) == pes[0]
        pes[0].consume_privacy_budget(eps)

    elif len(pes) == 1 and len(pes[0].parents) > 0:
        pes[0].consume_privacy_budget(eps)
        consume_privacy_budget(pes[0].parents, eps)

    else:
        # skip intermediate entities until all paths converge into a single entity
        # (lowest "single" common ancestor (LSCA) in a dag)
        consume_privacy_budget([get_lsca(pes)], eps)

def consumed_privacy_budget(name: str) -> float:
    return get_provenance_root(name).consumed_privacy_budget

def consumed_privacy_budget_all() -> dict[str, float]:
    global provenance_roots
    return {name: pe.consumed_privacy_budget for name, pe in provenance_roots.items()}

def get_lsca(pes: list[ProvenanceEntity]) -> ProvenanceEntity:
    max_depth = max([p.depth for p in pes])
    pe_set = set(pes)
    for d in reversed(range(max_depth + 1)):
        if len(pe_set) == 1:
            return next(iter(pe_set))
        pe_set = {pp for p in pe_set for pp in p.parents if p.depth == d}
    raise RuntimeError

# should not be exposed
def clear_global_states() -> None:
    global provenance_roots
    provenance_roots = {}



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import uuid
import privjail as pj

def new_sensitive_int(value: int) -> pj.SensitiveInt:
    # TODO: how to handle multimethod function types?
    return pj.SensitiveInt(value, distance=pj.Distance(1), root_name=str(uuid.uuid4())) + 0 # type: ignore[no-any-return]

def test_sensitive_real_number() -> None:
    x = new_sensitive_int(12)
    assert isinstance(x, pj.SensitiveInt)

    y = x * (1 / 12)
    assert isinstance(y, pj.SensitiveFloat)

    x = x + 1
    assert x._value == 13
    assert x.distance.max() == 1

    x = 1 + x
    assert x._value == 14
    assert x.distance.max() == 1

    x += 1
    assert x._value == 15
    assert x.distance.max() == 1

    x *= 3
    assert x._value == 45
    assert x.distance.max() == 3

    x = x * 2 + 1
    assert x._value == 91
    assert x.distance.max() == 6

    x -= 90
    assert x._value == 1
    assert x.distance.max() == 6

    with pytest.raises(TypeError):
        x * x

    z = x + y
    assert z._value == pytest.approx(2.0)
    assert z.distance.max() == pytest.approx(6.0 + 1 / 12)

    z = x - 2 * y
    assert z._value == pytest.approx(-1.0)
    assert z.distance.max() == pytest.approx(6.0 + 1 / 6)

    with pytest.raises(TypeError):
        x * y

def test_min_max() -> None:
    x = new_sensitive_int(11)
    y = x - 1
    z = x + 2

    assert pj.max(x, y)._value == 11
    assert pj.max(x, y).distance.max() == 1

    assert pj.min(x, y)._value == 10
    assert pj.min(x, y).distance.max() == 1

    assert pj.max(x, y, z)._value == 13
    assert pj.max(x, y, z).distance.max() == 1

    assert pj.min(x, y, z)._value == 10
    assert pj.min(x, y, z).distance.max() == 1

    z *= 2

    assert pj.max(x, y, z)._value == 26
    assert pj.max(x, y, z).distance.max() == 2

    assert pj.min(x, y, z)._value == 10
    assert pj.min(x, y, z).distance.max() == 2

    assert pj.max([x, y, z])._value == 26
    assert pj.max([x, y, z]).distance.max() == 2

    assert pj.min([x, y, z])._value == 10
    assert pj.min([x, y, z]).distance.max() == 2

    assert type(pj.max(x * 3.0, y, z)) == pj.SensitiveFloat
    assert pj.max(x * 3.0, y, z)._value == pytest.approx(33.0)
    assert pj.max(x * 3.0, y, z).distance.max() == pytest.approx(3.0)

    assert type(pj.min(x * 3.0, y, z)) == pj.SensitiveFloat
    assert pj.min(x * 3.0, y, z)._value == pytest.approx(10.0)
    assert pj.min(x * 3.0, y, z).distance.max() == pytest.approx(3.0)

    assert type(pj.max([x, y, z - 25.0])) == pj.SensitiveFloat
    assert pj.max([x, y, z - 25.0])._value == pytest.approx(11.0)
    assert pj.max([x, y, z - 25.0]).distance.max() == pytest.approx(2.0)

    assert type(pj.min([x, y, z - 25.0])) == pj.SensitiveFloat
    assert pj.min([x, y, z - 25.0])._value == pytest.approx(1.0)
    assert pj.min([x, y, z - 25.0]).distance.max() == pytest.approx(2.0)

    with pytest.raises(TypeError): pj.max()
    with pytest.raises(TypeError): pj.min()

    with pytest.raises(TypeError): pj.max(x)
    with pytest.raises(TypeError): pj.min(x)

    with pytest.raises(ValueError): pj.max([])
    with pytest.raises(ValueError): pj.min([])



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
import pytest
import importlib
import sys
from pathlib import Path

sys.path.append(str(Path(__file__).resolve().parent.parent / "src"))
provenance = importlib.import_module("privjail.provenance")

@pytest.fixture(autouse=True)
def setup() -> Any:
    provenance.clear_global_states()
    yield

def test_provenance_accumulation() -> None:
    pe0 = provenance.new_provenance_root("foo")
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 0

    provenance.consume_privacy_budget([pe0], 10)
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 10

    pe1 = provenance.new_provenance_node([pe0], "inclusive")
    pe2 = provenance.new_provenance_node([pe0], "inclusive")

    provenance.consume_privacy_budget([pe1], 20)
    provenance.consume_privacy_budget([pe2], 30)

    assert pe1.consumed_privacy_budget == 20
    assert pe2.consumed_privacy_budget == 30
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 60

    pe1e = provenance.new_provenance_node([pe1], "exclusive")
    pe1e1 = provenance.new_provenance_node([pe1e], "inclusive")
    pe1e2 = provenance.new_provenance_node([pe1e], "inclusive")

    provenance.consume_privacy_budget([pe1e1], 50)
    provenance.consume_privacy_budget([pe1e2], 30)

    assert pe1e1.consumed_privacy_budget == 50
    assert pe1e2.consumed_privacy_budget == 30
    assert pe1e.consumed_privacy_budget == 50
    assert pe1.consumed_privacy_budget == 70
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 110

    provenance.consume_privacy_budget([pe1e2], 30)

    assert pe1e1.consumed_privacy_budget == 50
    assert pe1e2.consumed_privacy_budget == 60
    assert pe1e.consumed_privacy_budget == 60
    assert pe1.consumed_privacy_budget == 80
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 120

    pe2e = provenance.new_provenance_node([pe2], "exclusive")
    pe2e1 = provenance.new_provenance_node([pe2e], "inclusive")
    pe2e2 = provenance.new_provenance_node([pe2e], "inclusive")
    pe2e3 = provenance.new_provenance_node([pe2e], "inclusive")
    pe2e1_2e2_1 = provenance.new_provenance_node([pe2e1, pe2e2], "inclusive")
    pe2e1_2e3_1 = provenance.new_provenance_node([pe2e1, pe2e3], "inclusive")
    pe2e2_2e3_1 = provenance.new_provenance_node([pe2e2, pe2e3], "inclusive")

    provenance.consume_privacy_budget([pe2e1_2e2_1], 10)
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 130
    provenance.consume_privacy_budget([pe2e2_2e3_1], 10)
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 140
    provenance.consume_privacy_budget([pe2e1_2e3_1], 10)
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 140

    provenance.consume_privacy_budget([pe2e3], 20)
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 160
    provenance.consume_privacy_budget([pe2e1_2e2_1], 10)
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 160
    provenance.consume_privacy_budget([pe2e2], 15)
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 165

    assert pe2e1.consumed_privacy_budget == 30
    assert pe2e2.consumed_privacy_budget == 45
    assert pe2e3.consumed_privacy_budget == 40

    pe21 = provenance.new_provenance_node([pe2], "inclusive")
    pe22 = provenance.new_provenance_node([pe2], "inclusive")
    pe21_22_1 = provenance.new_provenance_node([pe21, pe22], "inclusive")
    pe2e3__21_22_1__1 = provenance.new_provenance_node([pe2e3, pe21_22_1], "inclusive")

    provenance.consume_privacy_budget([pe2e3__21_22_1__1], 10)
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 175
    provenance.consume_privacy_budget([pe2e3__21_22_1__1], 10)
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 185

    pe0_ = provenance.new_provenance_root("bar")
    provenance.consume_privacy_budget([pe0_], 20)

    assert provenance.consumed_privacy_budget("bar") == pe0_.consumed_privacy_budget == 20
    assert provenance.consumed_privacy_budget("foo") == pe0.consumed_privacy_budget == 185



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import pandas
from .util import DPError
from .prisoner import Prisoner, SensitiveInt, SensitiveFloat, _max as max, _min as min, consumed_privacy_budget
from .distance import Distance
from .mechanism import laplace_mechanism, exponential_mechanism
from .egrpc import serve, connect, disconnect, proto_file_content

__all__ = [
    "pandas",
    "DPError",
    "Prisoner",
    "SensitiveInt",
    "SensitiveFloat",
    "max",
    "min",
    "Distance",
    "consumed_privacy_budget",
    "laplace_mechanism",
    "exponential_mechanism",
    "serve",
    "connect",
    "disconnect",
    "proto_file_content",
]



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import get_type_hints, get_origin, get_args, Union, List, Tuple, Dict, Any, TypeVar, Callable, Type, ParamSpec
from types import UnionType
from collections.abc import Sequence, Mapping
import inspect

# TODO: make egrpc independent of numpy
import numpy as _np

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")

TypeHint = Any

# https://github.com/python/mypy/issues/15630
def my_get_origin(type_hint: Any) -> Any:
    return get_origin(type_hint)

def is_type_match(obj: Any, type_hint: TypeHint) -> bool:
    type_origin = my_get_origin(type_hint)
    type_args = get_args(type_hint)

    if type_origin is None:
        return isinstance(obj, type_hint)

    elif type_origin in (Union, UnionType):
        return any(is_type_match(obj, th) for th in type_args)

    elif type_origin in (tuple, Tuple):
        return (
            isinstance(obj, tuple)
            and len(obj) == len(type_args)
            and all(is_type_match(o, th) for o, th in zip(obj, type_args))
        )

    elif type_origin in (list, List, Sequence):
        return (
            isinstance(obj, list)
            and all(is_type_match(o, type_args[0]) for o in obj)
        )

    elif type_origin in (dict, Dict, Mapping):
        return (
            isinstance(obj, dict)
            and all(is_type_match(k, type_args[0]) for k in obj.keys())
            and all(is_type_match(v, type_args[1]) for v in obj.values())
        )

    elif type_origin in (_np.integer, _np.floating):
        # TODO: consider type args (T in np.integer[T])
        return isinstance(obj, (_np.integer, _np.floating))

    else:
        raise TypeError(f"Type {type_origin} is not supported.")

def get_function_typed_params(func: Callable[P, R]) -> dict[str, TypeHint]:
    type_hints = get_type_hints(func)
    param_names = list(inspect.signature(func).parameters.keys())
    return {param_name: type_hints[param_name] for param_name in param_names}

def get_function_return_type(func: Callable[P, R]) -> TypeHint:
    type_hints = get_type_hints(func)
    return type_hints["return"]

def get_class_typed_members(cls: Type[T]) -> dict[str, TypeHint]:
    return get_type_hints(cls)

def get_method_typed_params(cls: Type[T], method: Callable[P, R]) -> dict[str, TypeHint]:
    type_hints = get_type_hints(method)
    param_names = list(inspect.signature(method).parameters.keys())
    return {param_names[0]: cls,
            **{param_name: type_hints[param_name] for param_name in param_names[1:]}}

def get_method_return_type(cls: Type[T], method: Callable[P, R]) -> TypeHint:
    type_hints = get_type_hints(method)
    return type_hints["return"]

def normalize_args(func: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> dict[str, Any]:
    sig = inspect.signature(func)
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    return bound_args.arguments



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import overload, TypeVar, Any, Literal, Iterator, Generic, Sequence, Mapping, TYPE_CHECKING
from abc import ABC, abstractmethod
import warnings
import json
import copy
from dataclasses import dataclass, field

import numpy as _np
import pandas as _pd

from .util import DPError, is_realnum, realnum
from .prisoner import Prisoner, SensitiveInt, SensitiveFloat, _max as smax, _min as smin
from .distance import Distance
from . import egrpc

T = TypeVar("T")

ElementType = realnum | str | bool

PTag = int

ptag_count = 0

def new_ptag() -> PTag:
    global ptag_count
    ptag_count += 1
    return ptag_count

def normalize_column_schema(col_schema: dict[str, Any]) -> dict[str, Any]:
    if "type" not in col_schema:
        raise ValueError("Column schema must have the 'type' field.")

    col_type = col_schema["type"]
    new_col_schema = dict(type=col_type, na_value=col_schema.get("na_value", ""))

    if col_type in ["int64", "Int64", "float64", "Float64"]:
        if "range" not in col_schema:
            new_col_schema["range"] = [None, None]

        else:
            expected_type = int if col_type in ["int64", "Int64"] else float
            if not isinstance(col_schema["range"], list) or len(col_schema["range"]) != 2 or \
                    (col_schema["range"][0] is not None and not isinstance(col_schema["range"][0], expected_type)) or \
                    (col_schema["range"][1] is not None and not isinstance(col_schema["range"][1], expected_type)):
                raise ValueError(f"The 'range' field must be in the form [a, b], where a and b is {expected_type} or null.")

            new_col_schema["range"] = col_schema["range"]

    elif col_type == "category":
        if "categories" not in col_schema:
            raise ValueError("Please specify the 'categories' field for the column of type 'category'.")

        if not isinstance(col_schema["categories"], list) or not all(isinstance(x, str) for x in col_schema["categories"]):
            raise ValueError("The 'categories' field must be a list of strings.")

        if len(col_schema["categories"]) == 0:
            raise ValueError("The 'categories' list must have at least one element.")

        new_col_schema["categories"] = col_schema["categories"]

    elif col_type == "string":
        pass

    else:
        raise ValueError(f"Type '{col_type}' is not supported for dataframe column types.")

    return new_col_schema

def apply_column_schema(ser: _pd.Series[Any], col_schema: dict[str, Any], col_name: str) -> _pd.Series[Any]:
    ser = ser.replace(col_schema["na_value"], None)

    if col_schema["type"] == "int64":
        try:
            return ser.astype("int64")
        except _pd.errors.IntCastingNaNError:
            raise ValueError(f"Column '{col_name}' may include NaN or inf values. Consider specifying 'Int64' for the column type.")

    elif col_schema["type"] == "category":
        category_dtype = _pd.api.types.CategoricalDtype(categories=col_schema["categories"])
        return ser.astype(category_dtype)

    else:
        return ser.astype(col_schema["type"]) # type: ignore[no-any-return]

def column_schema2domain(col_schema: dict[str, Any]) -> Domain:
    col_type = col_schema["type"]

    if col_type in ["int64", "Int64", "float64", "Float64"]:
        return RealDomain(dtype=col_type, range=col_schema["range"])

    elif col_type == "category":
        return CategoryDomain(categories=col_schema["categories"])

    elif col_type == "string":
        return StrDomain()

    else:
        raise RuntimeError

# suppress mypy errors
if TYPE_CHECKING:
    @dataclass
    class Domain(ABC):
        dtype: str

        @abstractmethod
        def type(self) -> type:
            pass

    @dataclass
    class BoolDomain(Domain):
        dtype: str = "bool"

        def type(self) -> type:
            return bool

    @dataclass
    class RealDomain(Domain):
        range: tuple[realnum | None, realnum | None]

        def type(self) -> Any:
            return int if self.dtype in ("int64", "Int64") else float

    @dataclass
    class StrDomain(Domain):
        dtype: str = "string"

        def type(self) -> type:
            return str

    @dataclass
    class CategoryDomain(Domain):
        dtype: str = "categories"
        categories: list[ElementType] = field(default_factory=list)

        def type(self) -> type:
            return str

    @dataclass
    class PrivDataFrameGroupBy:
        # TODO: groups are ordered?
        groups: Mapping[ElementType, PrivDataFrame]

        def __len__(self) -> int:
            return len(self.groups)

        def __iter__(self) -> Iterator[tuple[Any, PrivDataFrame]]:
            return iter(self.groups.items())

        def get_group(self, key: Any) -> PrivDataFrame:
            return self.groups[key]

else:
    @egrpc.dataclass
    class Domain(ABC):
        dtype: str

        @abstractmethod
        def type(self) -> type:
            pass

    @egrpc.dataclass
    class BoolDomain(Domain):
        dtype: str = "bool"

        def type(self) -> type:
            return bool

    @egrpc.dataclass
    class RealDomain(Domain):
        range: tuple[realnum | None, realnum | None]

        def type(self) -> Any:
            return int if self.dtype in ("int64", "Int64") else float

    @egrpc.dataclass
    class StrDomain(Domain):
        dtype: str = "string"

        def type(self) -> type:
            return str

    @egrpc.dataclass
    class CategoryDomain(Domain):
        dtype: str = "categories"
        categories: list[ElementType] = field(default_factory=list)

        def type(self) -> type:
            assert len(self.categories) > 0
            return type(self.categories[0]) # TODO: how about other elements?

    @egrpc.dataclass
    class PrivDataFrameGroupBy:
        # TODO: groups are ordered?
        groups: Mapping[ElementType, PrivDataFrame]

        def __len__(self) -> int:
            return len(self.groups)

        def __iter__(self) -> Iterator[tuple[Any, PrivDataFrame]]:
            return iter(self.groups.items())

        def get_group(self, key: Any) -> PrivDataFrame:
            return self.groups[key]

@egrpc.remoteclass
class PrivDataFrame(Prisoner[_pd.DataFrame]):
    """Private DataFrame.

    Each row in this dataframe object should have a one-to-one relationship with an individual (event-/row-/item-level DP).
    Therefore, the number of rows is treated as a sensitive value.
    """
    _ptag    : PTag
    _domains : Mapping[str, Domain]

    def __init__(self,
                 data         : Any,
                 domains      : Mapping[str, Domain],
                 distance     : Distance,
                 index        : Any                     = None,
                 columns      : Any                     = None,
                 dtype        : Any                     = None,
                 copy         : bool                    = False,
                 *,
                 parents      : Sequence[Prisoner[Any]] = [],
                 root_name    : str | None              = None,
                 preserve_row : bool | None             = None,
                 ):
        if len(parents) == 0:
            preserve_row = False
        elif preserve_row is None:
            raise ValueError("preserve_row is required when parents are specified.")

        self._domains = domains

        if preserve_row:
            parent_tags = [p._ptag for p in parents if isinstance(p, (PrivDataFrame, PrivSeries))]
            assert len(parent_tags) > 0 and all([parent_tags[0] == pt for pt in parent_tags])
            self._ptag = parent_tags[0]
        else:
            self._ptag = new_ptag()

        data = _pd.DataFrame(data, index, columns, dtype, copy)
        if preserve_row:
            super().__init__(value=data, distance=distance, parents=parents, root_name=root_name)
        else:
            super().__init__(value=data, distance=distance, parents=parents, root_name=root_name)

    def __len__(self) -> int:
        # We cannot return Prisoner() here because len() must be an integer value
        raise DPError("len(df) is not supported. Use df.shape[0] instead.")

    @egrpc.multimethod
    def __getitem__(self, key: str) -> PrivSeries[ElementType]:
        # TODO: consider duplicated column names
        value_type = self.domains[key].type()
        # TODO: how to pass `value_type` from server to client via egrpc?
        return PrivSeries[value_type](data         = self._value.__getitem__(key), # type: ignore[valid-type]
                                      domain       = self.domains[key],
                                      distance     = self.distance,
                                      parents      = [self],
                                      preserve_row = True)

    @__getitem__.register
    def _(self, key: list[str]) -> PrivDataFrame:
        new_domains = {c: d for c, d in self.domains.items() if c in key}
        return PrivDataFrame(data         = self._value.__getitem__(key),
                             domains      = new_domains,
                             distance     = self.distance,
                             parents      = [self],
                             preserve_row = True)

    @__getitem__.register
    def _(self, key: PrivSeries[bool]) -> PrivDataFrame:
        if self._ptag != key._ptag:
            raise DPError("df[bool_vec] cannot be accepted when df and bool_vec are not row-preserved.")

        return PrivDataFrame(data         = self._value.__getitem__(key._value),
                             domains      = self.domains,
                             distance     = self.distance,
                             parents      = [self, key],
                             preserve_row = False)

    @egrpc.multimethod
    def __setitem__(self, key: str, value: ElementType) -> None:
        # TODO: consider domain transform
        self._value[key] = value

    @__setitem__.register
    def _(self, key: str, value: PrivSeries[Any]) -> None:
        new_domains = dict()
        for col, domain in self.domains.items():
            if col == key:
                new_domains[col] = value.domain
            else:
                new_domains[col] = domain
        self._domains = new_domains

        self._value[key] = value._value

    @__setitem__.register
    def _(self, key: list[str], value: ElementType) -> None:
        # TODO: consider domain transform
        self._value[key] = value

    @__setitem__.register
    def _(self, key: list[str], value: PrivDataFrame) -> None:
        # TODO: consider domain transform
        self._value[key] = value._value

    @__setitem__.register
    def _(self, key: PrivSeries[bool], value: ElementType) -> None:
        # TODO: consider domain transform
        if self._ptag != key._ptag:
            raise DPError("df[bool_vec] cannot be accepted when df and bool_vec are not row-preserved.")

        self._value[key._value] = value

    @__setitem__.register
    def _(self, key: PrivSeries[bool], value: list[ElementType]) -> None:
        # TODO: consider domain transform
        if self._ptag != key._ptag:
            raise DPError("df[bool_vec] cannot be accepted when df and bool_vec are not row-preserved.")

        self._value[key._value] = value

    @__setitem__.register
    def _(self, key: PrivSeries[bool], value: PrivDataFrame) -> None:
        # TODO: consider domain transform
        if self._ptag != key._ptag:
            raise DPError("df[bool_vec] cannot be accepted when df and bool_vec are not row-preserved.")

        self._value[key._value] = value._value

    @egrpc.multimethod
    def __eq__(self, other: PrivDataFrame) -> PrivDataFrame:
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive dataframes for comparison can be different.")

        return PrivDataFrame(data         = self._value == other._value,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self, other],
                             preserve_row = True)

    @__eq__.register
    def _(self, other: ElementType) -> PrivDataFrame:
        return PrivDataFrame(data         = self._value == other,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self],
                             preserve_row = True)

    @egrpc.multimethod
    def __ne__(self, other: PrivDataFrame) -> PrivDataFrame:
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive dataframes for comparison can be different.")

        return PrivDataFrame(data         = self._value != other._value,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self, other],
                             preserve_row = True)

    @__ne__.register
    def _(self, other: ElementType) -> PrivDataFrame:
        return PrivDataFrame(data         = self._value != other,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self],
                             preserve_row = True)

    @egrpc.multimethod
    def __lt__(self, other: PrivDataFrame) -> PrivDataFrame: # type: ignore
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive dataframes for comparison can be different.")

        return PrivDataFrame(data         = self._value < other._value,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self, other],
                             preserve_row = True)

    @__lt__.register
    def _(self, other: ElementType) -> PrivDataFrame:
        return PrivDataFrame(data         = self._value < other,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self],
                             preserve_row = True)

    @egrpc.multimethod
    def __le__(self, other: PrivDataFrame) -> PrivDataFrame: # type: ignore
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive dataframes for comparison can be different.")

        return PrivDataFrame(data         = self._value <= other._value,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self, other],
                             preserve_row = True)

    @__le__.register
    def _(self, other: ElementType) -> PrivDataFrame:
        return PrivDataFrame(data         = self._value <= other,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self],
                             preserve_row = True)

    @egrpc.multimethod
    def __gt__(self, other: PrivDataFrame) -> PrivDataFrame: # type: ignore
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive dataframes for comparison can be different.")

        return PrivDataFrame(data         = self._value > other._value,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self, other],
                             preserve_row = True)

    @__gt__.register
    def _(self, other: ElementType) -> PrivDataFrame:
        return PrivDataFrame(data         = self._value > other,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self],
                             preserve_row = True)

    @egrpc.multimethod
    def __ge__(self, other: PrivDataFrame) -> PrivDataFrame: # type: ignore
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive dataframes for comparison can be different.")

        return PrivDataFrame(data         = self._value >= other._value,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self, other],
                             preserve_row = True)

    @__ge__.register
    def _(self, other: ElementType) -> PrivDataFrame:
        return PrivDataFrame(data         = self._value >= other,
                             domains      = {c: BoolDomain() for c in self.domains},
                             distance     = self.distance,
                             parents      = [self],
                             preserve_row = True)

    @egrpc.property
    def shape(self) -> tuple[SensitiveInt, int]:
        nrows = SensitiveInt(value=self._value.shape[0], distance=self.distance, parents=[self])
        ncols = self._value.shape[1]
        return (nrows, ncols)

    @egrpc.property
    def size(self) -> SensitiveInt:
        return SensitiveInt(value=self._value.size, distance=self.distance * len(self._value.columns), parents=[self])

    # TODO: define privjail's own Index[T] type
    @property
    def columns(self) -> _pd.Index[str]:
        return _pd.Index(self._get_columns())

    @egrpc.method
    def _get_columns(self) -> list[str]:
        return list(self._value.columns)

    @property
    def dtypes(self) -> _pd.Series[Any]:
        return self._value.dtypes

    @egrpc.property
    def domains(self) -> Mapping[str, Domain]:
        return self._domains

    # TODO: add test
    @egrpc.method
    def head(self, n: int = 5) -> PrivDataFrame:
        return PrivDataFrame(data         = self._value.head(n),
                             domains      = self.domains,
                             distance     = self.distance * 2,
                             parents      = [self],
                             preserve_row = False)

    # TODO: add test
    @egrpc.method
    def tail(self, n: int = 5) -> PrivDataFrame:
        return PrivDataFrame(data         = self._value.tail(n),
                             domains      = self.domains,
                             distance     = self.distance * 2,
                             parents      = [self],
                             preserve_row = False)

    @overload
    def sort_values(self,
                    by        : str | list[str],
                    *,
                    ascending : bool = ...,
                    inplace   : Literal[True],
                    ) -> None: ...

    @overload
    def sort_values(self,
                    by        : str | list[str],
                    *,
                    ascending : bool = ...,
                    inplace   : Literal[False] = ...,
                    ) -> PrivDataFrame: ...

    # TODO: add test
    @egrpc.method
    def sort_values(self,
                    by        : str | list[str],
                    *,
                    ascending : bool = True,
                    inplace   : bool = False,
                    ) -> PrivDataFrame | None:
        if inplace:
            self._value.sort_values(by, ascending=ascending, inplace=inplace, kind="stable")
            self._ptag = new_ptag()
            return None
        else:
            return PrivDataFrame(data         = self._value.sort_values(by, ascending=ascending, inplace=inplace, kind="stable"),
                                 domains      = self.domains,
                                 distance     = self.distance,
                                 parents      = [self],
                                 preserve_row = False)

    @overload
    def replace(self,
                to_replace : ElementType | None = ...,
                value      : ElementType | None = ...,
                *,
                inplace    : Literal[True],
                ) -> None: ...

    @overload
    def replace(self,
                to_replace : ElementType | None = ...,
                value      : ElementType | None = ...,
                *,
                inplace    : Literal[False] = ...,
                ) -> PrivDataFrame: ...

    @egrpc.method
    def replace(self,
                to_replace : ElementType | None = None,
                value      : ElementType | None = None,
                *,
                inplace    : bool = False,
                ) -> PrivDataFrame | None:
        if (not is_realnum(to_replace)) or (not is_realnum(value)):
            # TODO: consider string and category dtype
            raise NotImplementedError

        new_domains = dict()
        for col, domain in self.domains.items():
            if domain.dtype == "int64" and _np.isnan(value):
                new_domain = copy.copy(domain)
                new_domain.dtype = "Int64"
                new_domains[col] = new_domain

            elif isinstance(domain, RealDomain):
                a, b = domain.range
                if (a is None or a <= to_replace) and (b is None or to_replace <= b):
                    new_a = min(a, value) if a is not None else None # type: ignore[type-var]
                    new_b = max(b, value) if b is not None else None # type: ignore[type-var]

                    new_domain = copy.copy(domain)
                    new_domain.range = (new_a, new_b)
                    new_domains[col] = new_domain

                else:
                    new_domains[col] = domain

            else:
                new_domains[col] = domain

        if inplace:
            self._value.replace(to_replace, value, inplace=inplace) # type: ignore[arg-type]
            self._domains = new_domains
            return None
        else:
            return PrivDataFrame(data         = self._value.replace(to_replace, value, inplace=inplace), # type: ignore[arg-type]
                                 domains      = new_domains,
                                 distance     = self.distance,
                                 parents      = [self],
                                 preserve_row = True)

    @overload
    def dropna(self,
               *,
               inplace      : Literal[True],
               ignore_index : bool = ...,
               ) -> None: ...

    @overload
    def dropna(self,
               *,
               inplace      : Literal[False] = ...,
               ignore_index : bool = ...,
               ) -> PrivDataFrame: ...

    @egrpc.method
    def dropna(self,
               *,
               inplace      : bool = False,
               ignore_index : bool = False,
               ) -> PrivDataFrame | None:
        if ignore_index:
            raise DPError("`ignore_index` must be False. Index cannot be reindexed with positions.")

        new_domains = dict()
        for col, domain in self.domains.items():
            if domain.dtype == "Int64":
                new_domain = copy.copy(domain)
                new_domain.dtype = "int64"
                new_domains[col] = new_domain
            else:
                new_domains[col] = domain

        if inplace:
            self._value.dropna(inplace=inplace)
            self._domains = new_domains
            return None
        else:
            return PrivDataFrame(data         = self._value.dropna(inplace=inplace),
                                 domains      = new_domains,
                                 distance     = self.distance,
                                 parents      = [self],
                                 preserve_row = True)

    @egrpc.method
    def groupby(self,
                by         : str, # TODO: support more
                level      : int | None                   = None, # TODO: support multiindex?
                as_index   : bool                         = True,
                sort       : bool                         = True,
                group_keys : bool                         = True,
                observed   : bool                         = True,
                dropna     : bool                         = True,
                keys       : Sequence[ElementType] | None = None, # extra argument for privjail
                ) -> PrivDataFrameGroupBy:
        key_domain = self.domains[by]
        if isinstance(key_domain, CategoryDomain):
            keys = key_domain.categories

        if keys is None:
            raise DPError("Please provide the `keys` argument to prevent privacy leakage for non-categorical columns.")

        # TODO: consider extra arguments
        groupby_obj = self._value.groupby(by, observed=observed)

        # exclude groups not included in the specified `keys`
        groups = {key: df for key, df in groupby_obj if key in keys}

        # include empty groups for absent `keys` and sort by `keys`
        columns = self._value.columns
        dtypes = self._value.dtypes
        groups = {key: groups.get(key, _pd.DataFrame({c: _pd.Series(dtype=d) for c, d in zip(columns, dtypes)})) for key in keys} # type: ignore

        # create new child distance variables to express exclusiveness
        distances = self.distance.create_exclusive_distances(len(groups))

        # create a dummy prisoner to track exclusive provenance
        prisoner_dummy = Prisoner(0, self.distance, parents=[self], children_type="exclusive")

        # wrap each group by PrivDataFrame
        # TODO: update childrens' category domain that is chosen for the groupby key
        priv_groups = {key: PrivDataFrame(df, domains=self.domains, distance=d, parents=[prisoner_dummy], preserve_row=False) for (key, df), d in zip(groups.items(), distances)}

        return PrivDataFrameGroupBy(priv_groups) # type: ignore[arg-type]

# to avoid TypeError: type 'Series' is not subscriptable
# class PrivSeries(Prisoner[_pd.Series[T]]):
@egrpc.remoteclass
class PrivSeries(Generic[T], Prisoner[_pd.Series]): # type: ignore[type-arg]
    """Private Series.

    Each value in this series object should have a one-to-one relationship with an individual (event-/row-/item-level DP).
    Therefore, the number of values is treated as a sensitive value.
    """
    _domain : Domain
    _ptag   : PTag

    def __init__(self,
                 data         : Any,
                 domain       : Domain,
                 distance     : Distance,
                 index        : Any                     = None,
                 *,
                 parents      : Sequence[Prisoner[Any]] = [],
                 root_name    : str | None              = None,
                 preserve_row : bool | None             = None,
                 **kwargs    : Any,
                 ):
        if len(parents) == 0:
            preserve_row = False
        elif preserve_row is None:
            raise ValueError("preserve_row is required when parents are specified.")

        self._domain = domain

        if preserve_row:
            parent_tags = [p._ptag for p in parents if isinstance(p, (PrivDataFrame, PrivSeries))]
            assert len(parent_tags) > 0 and all([parent_tags[0] == pt for pt in parent_tags])
            self._ptag = parent_tags[0]
        else:
            self._ptag = new_ptag()

        data = _pd.Series(data, index, **kwargs)
        super().__init__(value=data, distance=distance, parents=parents, root_name=root_name)

    def __len__(self) -> int:
        # We cannot return Prisoner() here because len() must be an integer value
        raise DPError("len(ser) is not supported. Use ser.shape[0] or ser.size instead.")

    @egrpc.multimethod
    def __getitem__(self, key: PrivSeries[bool]) -> PrivSeries[T]:
        if self._ptag != key._ptag:
            raise DPError("ser[bool_vec] cannot be accepted when ser and bool_vec are not row-preserved.")

        return PrivSeries[T](data         = self._value.__getitem__(key._value),
                             domain       = self.domain,
                             distance     = self.distance,
                             parents      = [self],
                             preserve_row = False)

    @egrpc.multimethod
    def __setitem__(self, key: PrivSeries[bool], value: ElementType) -> None:
        if self._ptag != key._ptag:
            raise DPError("df[bool_vec] cannot be accepted when ser and bool_vec are not row-preserved.")

        self._value[key._value] = value

    @egrpc.multimethod
    def __eq__(self, other: PrivSeries[ElementType]) -> PrivSeries[bool]:
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive series for comparison can be different.")

        return PrivSeries[bool](data         = self._value == other._value,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self, other],
                                preserve_row = True)

    @__eq__.register
    def _(self, other: ElementType) -> PrivSeries[bool]:
        return PrivSeries[bool](data         = self._value == other,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self],
                                preserve_row = True)

    @egrpc.multimethod
    def __ne__(self, other: PrivSeries[ElementType]) -> PrivSeries[bool]:
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive series for comparison can be different.")

        return PrivSeries[bool](data         = self._value != other._value,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self, other],
                                preserve_row = True)

    @__ne__.register
    def _(self, other: ElementType) -> PrivSeries[bool]:
        return PrivSeries[bool](data         = self._value != other,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self],
                                preserve_row = True)

    @egrpc.multimethod
    def __lt__(self, other: PrivSeries[ElementType]) -> PrivSeries[bool]: # type: ignore
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive series for comparison can be different.")

        return PrivSeries[bool](data         = self._value < other._value,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self, other],
                                preserve_row = True)

    @__lt__.register
    def _(self, other: ElementType) -> PrivSeries[bool]:
        return PrivSeries[bool](data         = self._value < other,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self],
                                preserve_row = True)

    @egrpc.multimethod
    def __le__(self, other: PrivSeries[ElementType]) -> PrivSeries[bool]: # type: ignore
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive series for comparison can be different.")

        return PrivSeries[bool](data         = self._value <= other._value,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self, other],
                                preserve_row = True)

    @__le__.register
    def _(self, other: ElementType) -> PrivSeries[bool]:
        return PrivSeries[bool](data         = self._value <= other,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self],
                                preserve_row = True)

    @egrpc.multimethod
    def __gt__(self, other: PrivSeries[ElementType]) -> PrivSeries[bool]: # type: ignore
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive series for comparison can be different.")

        return PrivSeries[bool](data         = self._value > other._value,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self, other],
                                preserve_row = True)

    @__gt__.register
    def _(self, other: ElementType) -> PrivSeries[bool]:
        return PrivSeries[bool](data         = self._value > other,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self],
                                preserve_row = True)

    @egrpc.multimethod
    def __ge__(self, other: PrivSeries[ElementType]) -> PrivSeries[bool]: # type: ignore
        if self._ptag != other._ptag:
            raise DPError("Length of sensitive series for comparison can be different.")

        return PrivSeries[bool](data         = self._value >= other._value,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self, other],
                                preserve_row = True)

    @__ge__.register
    def _(self, other: ElementType) -> PrivSeries[bool]:
        return PrivSeries[bool](data         = self._value >= other,
                                domain       = BoolDomain(),
                                distance     = self.distance,
                                parents      = [self],
                                preserve_row = True)

    @egrpc.property
    def shape(self) -> tuple[SensitiveInt]:
        nrows = SensitiveInt(value=self._value.shape[0], distance=self.distance, parents=[self])
        return (nrows,)

    @egrpc.property
    def size(self) -> SensitiveInt:
        return SensitiveInt(value=self._value.size, distance=self.distance, parents=[self])

    @property
    def dtypes(self) -> Any:
        return self._value.dtypes

    @egrpc.property
    def domain(self) -> Domain:
        return self._domain

    # TODO: add test
    @egrpc.method
    def head(self, n: int = 5) -> PrivSeries[T]:
        return PrivSeries[T](data         = self._value.head(n),
                             domain       = self.domain,
                             distance     = self.distance * 2,
                             parents      = [self],
                             preserve_row = False)

    # TODO: add test
    @egrpc.method
    def tail(self, n: int = 5) -> PrivSeries[T]:
        return PrivSeries[T](data         = self._value.tail(n),
                             domain       = self.domain,
                             distance     = self.distance * 2,
                             parents      = [self],
                             preserve_row = False)

    @overload
    def sort_values(self,
                    *,
                    ascending : bool = ...,
                    inplace   : Literal[True],
                    ) -> None: ...

    @overload
    def sort_values(self,
                    *,
                    ascending : bool = ...,
                    inplace   : Literal[False] = ...,
                    ) -> PrivSeries[T]: ...

    # TODO: add test
    @egrpc.method
    def sort_values(self,
                    *,
                    ascending : bool = True,
                    inplace   : bool = False,
                    ) -> PrivSeries[T] | None:
        if inplace:
            self._value.sort_values(ascending=ascending, inplace=inplace, kind="stable")
            self._ptag = new_ptag()
            return None
        else:
            return PrivSeries[T](data         = self._value.sort_values(ascending=ascending, inplace=inplace, kind="stable"),
                                 domain       = self.domain,
                                 distance     = self.distance,
                                 parents      = [self],
                                 preserve_row = False)

    @overload
    def replace(self,
                to_replace : ElementType | None = ...,
                value      : ElementType | None = ...,
                *,
                inplace    : Literal[True],
                ) -> None: ...

    @overload
    def replace(self,
                to_replace : ElementType | None = ...,
                value      : ElementType | None = ...,
                *,
                inplace    : Literal[False] = ...,
                ) -> PrivSeries[T]: ...

    @egrpc.method
    def replace(self,
                to_replace : ElementType | None = None,
                value      : ElementType | None = None,
                *,
                inplace    : bool = False,
                ) -> PrivSeries[T] | None:
        if (not is_realnum(to_replace)) or (not is_realnum(value)):
            # TODO: consider string and category dtype
            raise NotImplementedError

        if self.domain.dtype == "int64" and _np.isnan(value):
            new_domain = copy.copy(self.domain)
            new_domain.dtype = "Int64"

        elif isinstance(self.domain, RealDomain):
            a, b = self.domain.range
            if (a is None or a <= to_replace) and (b is None or to_replace <= b):
                new_a = min(a, value) if a is not None else None # type: ignore[type-var]
                new_b = max(b, value) if b is not None else None # type: ignore[type-var]

                new_domain = copy.copy(self.domain)
                new_domain.range = (new_a, new_b)

            else:
                new_domain = self.domain

        else:
            new_domain = self.domain

        if inplace:
            self._value.replace(to_replace, value, inplace=inplace) # type: ignore[arg-type]
            self._domain = new_domain
            return None
        else:
            return PrivSeries[T](data         = self._value.replace(to_replace, value, inplace=inplace), # type: ignore[arg-type]
                                 domain       = new_domain,
                                 distance     = self.distance,
                                 parents      = [self],
                                 preserve_row = True)

    @overload
    def dropna(self,
               *,
               inplace      : Literal[True],
               ignore_index : bool = ...,
               ) -> None: ...

    @overload
    def dropna(self,
               *,
               inplace      : Literal[False] = ...,
               ignore_index : bool = ...,
               ) -> PrivSeries[T]: ...

    @egrpc.method
    def dropna(self,
               *,
               inplace      : bool = False,
               ignore_index : bool = False,
               ) -> PrivSeries[T] | None:
        if ignore_index:
            raise DPError("`ignore_index` must be False. Index cannot be reindexed with positions.")

        if self.domain.dtype == "Int64":
            new_domain = copy.copy(self.domain)
            new_domain.dtype = "int64"
        else:
            new_domain = self.domain

        if inplace:
            self._value.dropna(inplace=inplace)
            self._domain = new_domain
            return None
        else:
            return PrivSeries[T](data         = self._value.dropna(inplace=inplace),
                                 domain       = new_domain,
                                 distance     = self.distance,
                                 parents      = [self],
                                 preserve_row = True)

    @overload
    def clip(self,
             lower    : realnum | None = None,
             upper    : realnum | None = None,
             *,
             inplace  : Literal[True],
             ) -> None: ...

    @overload
    def clip(self,
             lower    : realnum | None = None,
             upper    : realnum | None = None,
             *,
             inplace  : Literal[False] = ...,
             ) -> PrivSeries[T]: ...

    @egrpc.method
    def clip(self,
             lower    : realnum | None = None,
             upper    : realnum | None = None,
             *,
             inplace  : bool = False,
             ) -> PrivSeries[T] | None:
        if not isinstance(self.domain, RealDomain):
            raise TypeError("Domain must be real numbers.")

        new_domain = copy.copy(self.domain)
        a, b = self.domain.range
        new_a = a if lower is None else lower if a is None else max(a, lower) # type: ignore[type-var]
        new_b = b if upper is None else upper if b is None else min(b, upper) # type: ignore[type-var]
        new_domain.range = (new_a, new_b)

        if inplace:
            self._value.clip(lower, upper, inplace=inplace) # type: ignore[arg-type]
            self._domain = new_domain
            return None
        else:
            return PrivSeries[T](data         = self._value.clip(lower, upper, inplace=inplace), # type: ignore[arg-type]
                                 domain       = new_domain,
                                 distance     = self.distance,
                                 parents      = [self],
                                 preserve_row = True)

    @egrpc.method
    def sum(self) -> SensitiveInt | SensitiveFloat:
        if not isinstance(self.domain, RealDomain):
            raise TypeError("Domain must be real numbers.")

        if None in self.domain.range:
            raise DPError("The range is unbounded. Use clip().")

        a, b = self.domain.range

        if a is None or b is None:
            raise DPError("The range is unbounded. Use clip().")

        new_distance = self.distance * max(_np.abs(a), _np.abs(b))

        s = self._value.sum()

        if self.domain.dtype in ["int64", "Int64"]:
            return SensitiveInt(s, new_distance, parents=[self])
        elif self.domain.dtype in ["float64", "Float64"]:
            return SensitiveFloat(s, new_distance, parents=[self])
        else:
            raise ValueError

    @egrpc.method
    def mean(self, eps: float) -> float:
        if not isinstance(self.domain, RealDomain):
            raise TypeError("Domain must be real numbers.")

        if eps <= 0:
            raise DPError(f"Invalid epsilon ({eps})")

        a, b = self.domain.range

        if a is None or b is None:
            raise DPError("The range is unbounded. Use clip().")

        sum_sensitivity = (self.distance * max(_np.abs(a), _np.abs(b))).max()
        count_sensitivity = self.distance.max()

        self.consume_privacy_budget(eps)

        s = _np.random.laplace(loc=float(self._value.sum()), scale=float(sum_sensitivity) / (eps / 2))
        c = _np.random.laplace(loc=float(self._value.shape[0]), scale=float(count_sensitivity) / (eps / 2))

        return s / c

    def value_counts(self,
                     normalize : bool                     = False,
                     sort      : bool                     = True,
                     ascending : bool                     = False,
                     bins      : int | None               = None,
                     dropna    : bool                     = True,
                     values    : list[ElementType] | None = None, # extra argument for privjail
                     ) -> SensitiveSeries[int]:
        # TODO: make SensitiveSeries a dataclass
        result = self._value_counts_impl(normalize, sort, ascending, bins, dropna, values)
        return SensitiveSeries[int](data=list(result.values()), index=list(result.keys()), dtype="object")

    @egrpc.method
    def _value_counts_impl(self,
                           normalize : bool                     = False,
                           sort      : bool                     = True,
                           ascending : bool                     = False,
                           bins      : int | None               = None,
                           dropna    : bool                     = True,
                           values    : list[ElementType] | None = None, # extra argument for privjail
                           ) -> dict[ElementType, SensitiveInt]:
        if normalize:
            # TODO: what is the sensitivity?
            raise NotImplementedError

        if bins is not None:
            # TODO: support continuous values
            raise NotImplementedError

        if sort:
            raise DPError("The `sort` argument must be False.")

        if isinstance(self.domain, CategoryDomain):
            values = self.domain.categories

        if values is None:
            raise DPError("Please provide the `values` argument to prevent privacy leakage.")

        if isinstance(values, Prisoner):
            raise DPError("`values` cannot be sensitive values.")

        if not dropna and not any(_np.isnan(values)): # type: ignore
            # TODO: consider handling for pd.NA
            warnings.warn("Counts for NaN will be dropped from the result because NaN is not included in `values`", UserWarning)

        counts = self._value.value_counts(normalize, sort, ascending, bins, dropna)

        # Select only the specified values and fill non-existent counts with 0
        counts = counts.reindex(values).fillna(0).astype(int)

        distances = self.distance.create_exclusive_distances(counts.size)

        prisoner_dummy = Prisoner(0, self.distance, parents=[self], children_type="exclusive")

        return {k: SensitiveInt(counts.loc[k], distance=distances[i], parents=[prisoner_dummy])
                for i, k in enumerate(counts.index)}

class SensitiveDataFrame(_pd.DataFrame):
    """Sensitive DataFrame.

    Each value in this dataframe object is considered a sensitive value.
    The numbers of rows and columns are not sensitive.
    This is typically created by counting queries like `pandas.crosstab()` and `pandas.pivot_table()`.
    """
    pass

# to avoid TypeError: type 'Series' is not subscriptable
# class SensitiveSeries(_pd.Series[T]):
class SensitiveSeries(Generic[T], _pd.Series): # type: ignore[type-arg]
    """Sensitive Series.

    Each value in this series object is considered a sensitive value.
    The numbers of values are not sensitive.
    This is typically created by counting queries like `PrivSeries.value_counts()`.
    """

    def max(self, *args: Any, **kwargs: Any) -> SensitiveInt | SensitiveFloat:
        # TODO: args?
        return smax(self)

    def min(self, *args: Any, **kwargs: Any) -> SensitiveInt | SensitiveFloat:
        # TODO: args?
        return smin(self)

@egrpc.function
def read_csv(filepath: str, schemapath: str | None = None) -> PrivDataFrame:
    # TODO: more vaildation for the input data
    df = _pd.read_csv(filepath)

    if schemapath is not None:
        with open(schemapath, "r") as f:
            schema = json.load(f)
    else:
        schema = dict()

    domains = dict()
    for col in df.columns:
        if not isinstance(col, str):
            raise ValueError("Column name must be a string.")

        if col in schema:
            col_schema = schema[col]
        else:
            col_schema = dict(type="string" if df.dtypes[col] == "object" else df.dtypes[col])

        col_schema = normalize_column_schema(col_schema)

        df[col] = apply_column_schema(df[col], col_schema, col)

        domains[col] = column_schema2domain(col_schema)

    return PrivDataFrame(data=df, domains=domains, distance=Distance(1), root_name=filepath)

def crosstab(index        : PrivSeries[Any] | list[PrivSeries[Any]],
             columns      : PrivSeries[Any] | list[PrivSeries[Any]],
             values       : PrivSeries[Any] | None                          = None,
             rownames     : list[str] | None                                = None,
             colnames     : list[str] | None                                = None,
             rowvalues    : list[Any] | _pd.Series[Any] | None              = None, # extra argument for privjail
             colvalues    : list[Any] | _pd.Series[Any] | None              = None, # extra argument for privjail
             *,
             aggfunc      : Any                                             = None,
             margins      : bool                                            = False,
             margins_name : str                                             = "All",
             dropna       : bool                                            = True,
             normalize    : bool | Literal["all", "index", "columns", 0, 1] = False,
             ) -> SensitiveDataFrame:
    if normalize is not False:
        # TODO: what is the sensitivity?
        raise NotImplementedError

    if values is not None or aggfunc is not None:
        # TODO: hard to accept arbitrary user functions
        raise NotImplementedError

    if margins:
        # Sensitivity must be managed separately for value counts and margins
        raise DPError("`margins=True` is not supported. Please manually calculate margins after adding noise.")

    if rowvalues is None or colvalues is None:
        # TODO: track the value domain to automatically determine the output dimension
        raise DPError("Please provide the `rowvalues`/`colvalues` arguments to prevent privacy leakage.")

    if not dropna and (not any(_np.isnan(rowvalues)) or not any(_np.isnan(colvalues))):
        # TODO: consider handling for pd.NA
        warnings.warn("Counts for NaN will be dropped from the result because NaN is not included in `rowvalues`/`colvalues`", UserWarning)

    if isinstance(index, list):
        if len(index) == 0:
            raise ValueError("Empty list is passed to `index`.")

        # TODO: support list of series (need to accept multiindex for rowvalues/colvalues)
        raise NotImplementedError
    else:
        index = [index]

    if isinstance(columns, list):
        if len(columns) == 0:
            raise ValueError("Empty list is passed to `columns`.")

        # TODO: support list of series (need to accept multiindex for rowvalues/colvalues)
        raise NotImplementedError
    else:
        columns = [columns]

    if not all(index[0]._ptag == x._ptag for x in index + columns):
        # TODO: reconsider whether this should be disallowed, as length difference does not cause an error
        raise DPError("Series in `index` and `columns` must have the same length.")

    counts = _pd.crosstab([x._value for x in index], [x._value for x in columns],
                          values=None, rownames=rownames, colnames=colnames,
                          aggfunc=None, margins=False, margins_name=margins_name,
                          dropna=dropna, normalize=False)

    # Select only the specified values and fill non-existent counts with 0
    counts = counts.reindex(rowvalues, axis="index").reindex(colvalues, axis="columns").fillna(0).astype(int)

    distances = index[0].distance.create_exclusive_distances(counts.size)

    prisoner_dummy = Prisoner(0, index[0].distance, parents=index + columns, children_type="exclusive")

    priv_counts = SensitiveDataFrame(index=counts.index, columns=counts.columns)
    i = 0
    for idx in counts.index:
        for col in counts.columns:
            priv_counts.loc[idx, col] = SensitiveInt(counts.loc[idx, col], distance=distances[i], parents=[prisoner_dummy]) # type: ignore
            i += 1

    return priv_counts

# TODO: change multifunction -> function by type checking in egrpc.function
@egrpc.multifunction
def cut(x              : PrivSeries[Any],
        bins           : list[int] | list[float],
        right          : bool                            = True,
        labels         : list[ElementType] | bool | None = None,
        retbins        : bool                            = False,
        precision      : int                             = 3,
        include_lowest : bool                            = False
        # TODO: add more parameters
        ) -> PrivSeries[Any]:
    ser = _pd.cut(x._value, bins=bins, right=right, labels=labels, retbins=retbins, precision=precision, include_lowest=include_lowest) # type: ignore

    new_domain = CategoryDomain(categories=list(ser.dtype.categories))

    return PrivSeries[Any](ser, domain=new_domain, distance=x.distance, parents=[x], preserve_row=True)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar, Callable, Type, Any, ParamSpec
from types import ModuleType
from concurrent import futures
import traceback
import grpc # type: ignore[import-untyped]

from . import names
from .proto_interface import ProtoMsg

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")

dynamic_pb2_grpc: ModuleType | None = None

def init_grpc(module: ModuleType) -> None:
    global dynamic_pb2_grpc
    dynamic_pb2_grpc = module

def get_grpc_module() -> ModuleType:
    global dynamic_pb2_grpc
    assert dynamic_pb2_grpc is not None
    return dynamic_pb2_grpc

HandlerType = Callable[[Any, ProtoMsg, Any], ProtoMsg]

proto_handlers: dict[str, dict[str, HandlerType]] = {}

def grpc_register_function(func: Callable[P, R], handler: HandlerType) -> None:
    proto_service_name = names.proto_function_service_name(func)
    proto_rpc_name = names.proto_function_rpc_name(func)

    # for debugging
    def wrapper(self: Any, proto_req: ProtoMsg, context: Any) -> ProtoMsg:
        try:
            return handler(self, proto_req, context)
        except:
            traceback.print_exc()
            raise

    global proto_handlers
    assert proto_service_name not in proto_handlers
    # proto_handlers[proto_service_name] = {proto_rpc_name: handler}
    proto_handlers[proto_service_name] = {proto_rpc_name: wrapper}

def grpc_register_method(cls: Type[T], method: Callable[P, R], handler: HandlerType) -> None:
    proto_service_name = names.proto_remoteclass_service_name(cls)
    proto_rpc_name = names.proto_method_rpc_name(cls, method)

    # for debugging
    def wrapper(self: Any, proto_req: ProtoMsg, context: Any) -> ProtoMsg:
        try:
            return handler(self, proto_req, context)
        except:
            traceback.print_exc()
            raise

    global proto_handlers
    if proto_service_name not in proto_handlers:
        proto_handlers[proto_service_name] = {}
    # proto_handlers[proto_service_name][proto_rpc_name] = handler
    proto_handlers[proto_service_name][proto_rpc_name] = wrapper

def init_server(port: int) -> Any:
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1),
                         maximum_concurrent_rpcs=1,
                         options=(("grpc.so_reuseport", 0),))

    global proto_handlers
    for proto_service_name, handlers in proto_handlers.items():
        DynamicServicer = type(f"{proto_service_name}DynamicServicer",
                               (getattr(dynamic_pb2_grpc, f"{proto_service_name}Servicer"),),
                               handlers)

        add_service_fn = getattr(dynamic_pb2_grpc, f"add_{proto_service_name}Servicer_to_server")
        add_service_fn(DynamicServicer(), server)

    server.add_insecure_port(f"[::]:{port}")

    return server

ChannelType = Any

client_channel: ChannelType | None = None

def init_client(hostname: str, port: int) -> None:
    global client_channel
    client_channel = grpc.insecure_channel(f"{hostname}:{port}")

def del_client() -> None:
    global client_channel
    client_channel = None

def get_client_channel() -> ChannelType:
    global client_channel
    assert client_channel is not None
    return client_channel

def remote_connected() -> bool:
    return client_channel is not None

def grpc_function_call(func: Callable[P, R], proto_req: ProtoMsg) -> ProtoMsg:
    proto_service_name = names.proto_function_service_name(func)
    proto_rpc_name = names.proto_function_rpc_name(func)

    channel = get_client_channel()
    stub = getattr(get_grpc_module(), f"{proto_service_name}Stub")(channel)
    proto_res = getattr(stub, proto_rpc_name)(proto_req)

    return proto_res

def grpc_method_call(cls: Type[T], method: Callable[P, R], proto_req: ProtoMsg) -> ProtoMsg:
    proto_service_name = names.proto_remoteclass_service_name(cls)
    proto_rpc_name = names.proto_method_rpc_name(cls, method)

    channel = get_client_channel()
    stub = getattr(get_grpc_module(), f"{proto_service_name}Stub")(channel)
    proto_res = getattr(stub, proto_rpc_name)(proto_req)

    return proto_res



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, TypeVar, overload

import numpy as _np
import pandas as _pd

from .util import DPError, floating, realnum
from .prisoner import Prisoner, SensitiveInt, SensitiveFloat
from .pandas import SensitiveSeries, SensitiveDataFrame
from . import egrpc

T = TypeVar("T")

@egrpc.multifunction
def laplace_mechanism(prisoner: SensitiveInt | SensitiveFloat, eps: floating) -> float:
    sensitivity = prisoner.distance.max()

    if sensitivity <= 0:
        raise DPError(f"Invalid sensitivity ({sensitivity})")

    if eps <= 0:
        raise DPError(f"Invalid epsilon ({eps})")

    prisoner.consume_privacy_budget(float(eps))

    return float(_np.random.laplace(loc=prisoner._value, scale=sensitivity / eps))

@laplace_mechanism.register(remote=False) # type: ignore
def _(prisoner: SensitiveSeries[realnum], eps: floating) -> _pd.Series: # type: ignore[type-arg]
    return prisoner.map(lambda x: laplace_mechanism(x, eps))

@laplace_mechanism.register(remote=False) # type: ignore
def _(prisoner: SensitiveDataFrame, eps: floating) -> _pd.DataFrame:
    return prisoner.map(lambda x: laplace_mechanism(x, eps))

# TODO: add test
@overload
def exponential_mechanism(scores: list[SensitiveInt | SensitiveFloat], eps: float) -> int: ...
@overload
def exponential_mechanism(scores: dict[T, SensitiveInt | SensitiveFloat], eps: float) -> T: ...

def exponential_mechanism(scores: list[SensitiveInt | SensitiveFloat] | dict[Any, SensitiveInt | SensitiveFloat], eps: float) -> Any:
    if isinstance(scores, dict):
        idx = _exponential_mechanism_impl(list(scores.values()), eps)
        return list(scores.keys())[idx]
    elif isinstance(scores, list):
        return _exponential_mechanism_impl(scores, eps)
    else:
        raise ValueError("scores must be a list or dict.")

@egrpc.function
def _exponential_mechanism_impl(scores: list[SensitiveInt | SensitiveFloat], eps: float) -> int:
    if len(scores) == 0:
        raise ValueError("scores must have at least one element.")

    sensitivity = max([v.distance.max() for v in scores]) # type: ignore[type-var]

    if sensitivity <= 0:
        raise DPError(f"Invalid sensitivity ({sensitivity})")

    if eps <= 0:
        raise DPError(f"Invalid epsilon ({eps})")

    # create a dummy prisoner to propagate budget consumption to all prisoners
    prisoner_dummy = Prisoner(0, scores[0].distance, parents=scores)
    prisoner_dummy.consume_privacy_budget(eps)

    exponents = [eps * s._value / sensitivity / 2 for s in scores]
    # to prevent too small or large values (-> 0 or inf)
    M: float = _np.max(exponents) # type: ignore[arg-type]
    p = [_np.exp(x - M) for x in exponents]
    p /= sum(p)
    return _np.random.choice(len(scores), p=p)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TypeVar, Generic, Any, overload, Iterable, cast, Sequence

import numpy as _np

from .util import integer, floating, realnum, is_integer, is_floating
from .provenance import ProvenanceEntity, new_provenance_root, new_provenance_node, consume_privacy_budget, consumed_privacy_budget_all, ChildrenType
from .distance import Distance, _max as dmax
from . import egrpc

T = TypeVar("T")

class Prisoner(Generic[T]):
    _value     : T
    distance   : Distance
    provenance : list[ProvenanceEntity]

    def __init__(self,
                 value         : T,
                 distance      : Distance,
                 *,
                 parents       : Sequence[Prisoner[Any]] = [],
                 root_name     : str | None              = None,
                 children_type : ChildrenType            = "inclusive",
                 ):
        self._value   = value
        self.distance = distance

        if distance.is_zero():
            self.provenance = []

        elif len(parents) == 0:
            if root_name is None:
                raise ValueError("Both parents and root_name are not specified.")

            self.provenance = [new_provenance_root(root_name)]

        elif children_type == "inclusive":
            parent_provenance = list({pe for p in parents for pe in p.provenance})

            if len(parents) == 1 and parents[0].provenance[0].children_type == "exclusive":
                self.provenance = [new_provenance_node(parent_provenance, "inclusive")]
            else:
                self.provenance = parent_provenance

        elif children_type == "exclusive":
            parent_provenance = list({pe for p in parents for pe in p.provenance})
            self.provenance = [new_provenance_node(parent_provenance, "exclusive")]

        else:
            raise RuntimeError

    def __str__(self) -> str:
        # return f"Prisoner({type(self._value)}, distance={self.distance.max()})"
        return f"Prisoner()"

    def __repr__(self) -> str:
        # return f"Prisoner({type(self._value)}, distance={self.distance.max()})"
        return f"Prisoner()"

    def consume_privacy_budget(self, privacy_budget: float) -> None:
        consume_privacy_budget(self.provenance, privacy_budget)

    def root_name(self) -> str:
        assert len(self.provenance) > 0
        return self.provenance[0].root_name

@egrpc.remoteclass
class SensitiveInt(Prisoner[integer]):
    def __init__(self,
                 value         : integer,
                 distance      : Distance                = Distance(0),
                 *,
                 parents       : Sequence[Prisoner[Any]] = [],
                 root_name     : str | None              = None,
                 children_type : ChildrenType            = "inclusive",
                 ):
        if not is_integer(value):
            raise ValueError("`value` must be int for SensitveInt.")
        super().__init__(value, distance, parents=parents, root_name=root_name, children_type=children_type)

    @egrpc.multimethod
    def __add__(self, other: integer) -> SensitiveInt:
        return SensitiveInt(self._value + other, distance=self.distance, parents=[self])

    @__add__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(self._value + other, distance=self.distance, parents=[self])

    @__add__.register
    def _(self, other: SensitiveInt) -> SensitiveInt:
        return SensitiveInt(self._value + other._value, distance=self.distance + other.distance, parents=[self, other])

    @__add__.register
    def _(self, other: SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(self._value + other._value, distance=self.distance + other.distance, parents=[self, other])

    @egrpc.multimethod
    def __radd__(self, other: integer) -> SensitiveInt: # type: ignore[misc]
        return SensitiveInt(other + self._value, distance=self.distance, parents=[self])

    @__radd__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(other + self._value, distance=self.distance, parents=[self])

    @__radd__.register
    def _(self, other: SensitiveInt) -> SensitiveInt:
        return SensitiveInt(other._value + self._value, distance=self.distance + other.distance, parents=[self, other])

    @__radd__.register
    def _(self, other: SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(other._value + self._value, distance=self.distance + other.distance, parents=[self, other])

    @egrpc.multimethod
    def __sub__(self, other: integer) -> SensitiveInt:
        return SensitiveInt(self._value - other, distance=self.distance, parents=[self])

    @__sub__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(self._value - other, distance=self.distance, parents=[self])

    @__sub__.register
    def _(self, other: SensitiveInt) -> SensitiveInt:
        return SensitiveInt(self._value - other._value, distance=self.distance + other.distance, parents=[self, other])

    @__sub__.register
    def _(self, other: SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(self._value - other._value, distance=self.distance + other.distance, parents=[self, other])

    @egrpc.multimethod
    def __rsub__(self, other: integer) -> SensitiveInt: # type: ignore[misc]
        return SensitiveInt(other - self._value, distance=self.distance, parents=[self])

    @__rsub__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(other - self._value, distance=self.distance, parents=[self])

    @__rsub__.register
    def _(self, other: SensitiveInt) -> SensitiveInt:
        return SensitiveInt(other._value - self._value, distance=self.distance + other.distance, parents=[self, other])

    @__rsub__.register
    def _(self, other: SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(other._value - self._value, distance=self.distance + other.distance, parents=[self, other])

    @egrpc.multimethod
    def __mul__(self, other: integer) -> SensitiveInt:
        return SensitiveInt(self._value * other, distance=self.distance * _np.abs(other), parents=[self])

    @__mul__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(self._value * other, distance=self.distance * _np.abs(other), parents=[self])

    @egrpc.multimethod
    def __rmul__(self, other: integer) -> SensitiveInt: # type: ignore[misc]
        return SensitiveInt(other * self._value, distance=self.distance * _np.abs(other), parents=[self])

    @__rmul__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(other * self._value, distance=self.distance * _np.abs(other), parents=[self])

@egrpc.remoteclass
class SensitiveFloat(Prisoner[floating]):
    def __init__(self,
                 value         : floating,
                 distance      : Distance                = Distance(0),
                 *,
                 parents       : Sequence[Prisoner[Any]] = [],
                 root_name     : str | None              = None,
                 children_type : ChildrenType            = "inclusive",
                 ):
        if not is_floating(value):
            raise ValueError("`value` must be float for SensitveFloat.")
        super().__init__(value, distance, parents=parents, root_name=root_name, children_type=children_type)

    @egrpc.multimethod
    def __add__(self, other: realnum) -> SensitiveFloat:
        return SensitiveFloat(self._value + other, distance=self.distance, parents=[self])

    @__add__.register
    def _(self, other: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(self._value + other._value, distance=self.distance + other.distance, parents=[self, other])

    @egrpc.multimethod
    def __radd__(self, other: realnum) -> SensitiveFloat: # type: ignore[misc]
        return SensitiveFloat(other + self._value, distance=self.distance, parents=[self])

    @__radd__.register
    def _(self, other: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(other._value + self._value, distance=self.distance + other.distance, parents=[self, other])

    @egrpc.multimethod
    def __sub__(self, other: realnum) -> SensitiveFloat:
        return SensitiveFloat(self._value - other, distance=self.distance, parents=[self])

    @__sub__.register
    def _(self, other: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(self._value - other._value, distance=self.distance + other.distance, parents=[self, other])

    @egrpc.multimethod
    def __rsub__(self, other: realnum) -> SensitiveFloat: # type: ignore[misc]
        return SensitiveFloat(other - self._value, distance=self.distance, parents=[self])

    @__rsub__.register
    def _(self, other: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(other._value - self._value, distance=self.distance + other.distance, parents=[self, other])

    @egrpc.multimethod
    def __mul__(self, other: realnum) -> SensitiveFloat:
        return SensitiveFloat(self._value * other, distance=self.distance * _np.abs(other), parents=[self])

    @egrpc.multimethod
    def __rmul__(self, other: realnum) -> SensitiveFloat: # type: ignore[misc]
        return SensitiveFloat(other * self._value, distance=self.distance * _np.abs(other), parents=[self])

@egrpc.multifunction
def max2(a: SensitiveInt | SensitiveFloat, b: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
    return SensitiveFloat(max(float(a._value), float(b._value)), distance=dmax(a.distance, b.distance), parents=[a, b])

@max2.register
def _(a: SensitiveInt, b: SensitiveInt) -> SensitiveInt:
    return SensitiveInt(max(int(a._value), int(b._value)), distance=dmax(a.distance, b.distance), parents=[a, b])

@overload
def _max(*args: SensitiveInt) -> SensitiveInt: ...
@overload
def _max(*args: SensitiveFloat) -> SensitiveFloat: ...
@overload
def _max(*args: Iterable[SensitiveInt]) -> SensitiveInt: ...
@overload
def _max(*args: Iterable[SensitiveFloat]) -> SensitiveFloat: ...
@overload
def _max(*args: Iterable[SensitiveInt | SensitiveFloat] | SensitiveInt | SensitiveFloat) -> SensitiveInt | SensitiveFloat: ...

def _max(*args: Iterable[SensitiveInt | SensitiveFloat] | SensitiveInt | SensitiveFloat) -> SensitiveInt | SensitiveFloat:
    if len(args) == 0:
        raise TypeError("max() expected at least one argment.")

    if len(args) == 1:
        if not isinstance(args[0], Iterable):
            raise TypeError("The first arg passed to max() is not iterable.")
        iterable = args[0]
    else:
        iterable = cast(tuple[SensitiveInt | SensitiveFloat, ...], args)

    it = iter(iterable)
    try:
        result = next(it)
    except StopIteration:
        raise ValueError("List passed to max() is empty.")

    for x in it:
        result = max2(result, x)

    return result

@egrpc.multifunction
def min2(a: SensitiveInt | SensitiveFloat, b: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
    return SensitiveFloat(min(float(a._value), float(b._value)), distance=dmax(a.distance, b.distance), parents=[a, b])

@min2.register
def _(a: SensitiveInt, b: SensitiveInt) -> SensitiveInt:
    return SensitiveInt(min(int(a._value), int(b._value)), distance=dmax(a.distance, b.distance), parents=[a, b])

@overload
def _min(*args: SensitiveInt) -> SensitiveInt: ...
@overload
def _min(*args: SensitiveFloat) -> SensitiveFloat: ...
@overload
def _min(*args: Iterable[SensitiveInt]) -> SensitiveInt: ...
@overload
def _min(*args: Iterable[SensitiveFloat]) -> SensitiveFloat: ...
@overload
def _min(*args: Iterable[SensitiveInt | SensitiveFloat] | SensitiveInt | SensitiveFloat) -> SensitiveInt | SensitiveFloat: ...

def _min(*args: Iterable[SensitiveInt | SensitiveFloat] | SensitiveInt | SensitiveFloat) -> SensitiveInt | SensitiveFloat:
    if len(args) == 0:
        raise TypeError("min() expected at least one argment.")

    if len(args) == 1:
        if not isinstance(args[0], Iterable):
            raise TypeError("The first arg passed to min() is not iterable.")
        iterable = args[0]
    else:
        iterable = cast(tuple[SensitiveInt | SensitiveFloat, ...], args)

    it = iter(iterable)
    try:
        result = next(it)
    except StopIteration:
        raise ValueError("List passed to min() is empty.")

    for x in it:
        result = min2(result, x)

    return result

@egrpc.function
def consumed_privacy_budget() -> dict[str, float]:
    return consumed_privacy_budget_all()



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import sys
from pathlib import Path

sys.path.append(str(Path(__file__).resolve().parent.parent / "src"))
distance = importlib.import_module("privjail.distance")

def test_distance() -> None:
    d = distance.Distance(1)
    assert d.max() == 1

    d = d * 2
    assert d.max() == 2

    d = d + 1
    assert d.max() == 3

    x = distance.new_distance_var()
    y = distance.new_distance_var()
    z = distance.new_distance_var()

    constraints = {distance.Constraint(frozenset({x, y, z}), 1)}

    dx = distance.Distance(x, constraints)
    dy = distance.Distance(y, constraints)
    dz = distance.Distance(z, constraints)

    assert dx.max() == 1

    d = dx + dy
    assert d.max() == 1

    d = d + dz
    assert d.max() == 1

    d = d * 2
    assert d.max() == 2

    d = d + 1
    assert d.max() == 3

    x_ = distance.new_distance_var()
    y_ = distance.new_distance_var()
    z_ = distance.new_distance_var()

    constraints |= {distance.Constraint(frozenset({x_, y_, z_}), x)}

    dx_ = distance.Distance(x_, constraints)
    dy_ = distance.Distance(y_, constraints)
    dz_ = distance.Distance(z_, constraints)

    assert dx_.max() == 1

    d = dx_ + dy_ + dz_
    assert d.max() == 1

    d = d * 4
    assert d.max() == 4

    d = d + dx
    assert d.max() == 5

    d = d + dy + dz
    assert d.max() == 5

    d = d * 2
    assert d.max() == 10

    d = d + dx_
    assert d.max() == 11



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import privjail as pj
from privjail import pandas as ppd

def main():
    df = ppd.read_csv("data/adult_train.csv", "schema/adult.json").dropna()

    n = 1000

    print()
    print("Income stats of younger age groups:")
    print(pj.laplace_mechanism(df.sort_values("age").head(n)["income"].value_counts(sort=False), eps=1.0))

    print()
    print("Income stats of older age groups:")
    print(pj.laplace_mechanism(df.sort_values("age").tail(n)["income"].value_counts(sort=False), eps=1.0))

    print()
    print("Average age by income:")
    for income, df_ in df.groupby("income"):
        mean_age = df_["age"].mean(eps=1.0)
        print(f"{income}: {mean_age}")

    print()
    print("Consumed Privacy Budget:")
    print(pj.consumed_privacy_budget())

if __name__ == "__main__":
    if len(sys.argv) < 2:
        mode = "local"
    else:
        mode = sys.argv[1]

    if mode in ("server", "client"):
        if len(sys.argv) < 4:
            raise ValueError(f"Usage: {sys.argv[0]} {mode} <host> <port>")

        host = sys.argv[2]
        port = int(sys.argv[3])

    if mode == "local":
        main()

    elif mode == "server":
        # print(pj.proto_file_content())
        pj.serve(port)

    elif mode == "client":
        pj.connect(host, port)
        main()

    else:
        raise ValueError(f"Usage: {sys.argv[0]} [local|server|client] [host] [port]")



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Any
import uuid
import pytest
import pandas as pd
import numpy as np
import privjail as pj
from privjail import pandas as ppd

def load_dataframe() -> tuple[ppd.PrivDataFrame, pd.DataFrame]:
    data = {
        "a": [1, 2, 3, 4, 5],
        "b": [2, 4, 4, 4, 3],
    }
    domains = {
        "a": ppd.RealDomain(dtype="int64", range=(None, None)),
        "b": ppd.RealDomain(dtype="int64", range=(None, None)),
    }
    pdf = ppd.PrivDataFrame(data, domains=domains, distance=pj.Distance(1), root_name=str(uuid.uuid4()))
    df = pd.DataFrame(data)
    assert (pdf.columns == df.columns).all()
    assert (pdf._value == df).all().all()
    return pdf, df

def assert_equal_sensitive_series(sensitive_ser: ppd.SensitiveSeries[Any], expected_ser: pd.Series[Any]) -> None:
    assert isinstance(sensitive_ser, ppd.SensitiveSeries)
    assert (sensitive_ser.index == expected_ser.index).all()
    for idx in sensitive_ser.index:
        assert sensitive_ser.loc[idx]._value == expected_ser.loc[idx]

def assert_equal_sensitive_dataframes(sensitive_df: ppd.SensitiveDataFrame, expected_df: pd.DataFrame) -> None:
    assert isinstance(sensitive_df, ppd.SensitiveDataFrame)
    assert (sensitive_df.index == expected_df.index).all()
    assert (sensitive_df.columns == expected_df.columns).all()
    for idx in sensitive_df.index:
        for col in sensitive_df.columns:
            assert sensitive_df.loc[idx, col]._value == expected_df.loc[idx, col]

def test_priv_dataframe_size() -> None:
    pdf, df = load_dataframe()

    # The `shape` member should be a pair of a sensitive value (row) and a non-sensitive value (column)
    assert isinstance(pdf.shape, tuple)
    assert len(pdf.shape) == len(df.shape) == 2
    assert isinstance(pdf.shape[0], pj.Prisoner)
    assert pdf.shape[0]._value == df.shape[0]
    assert pdf.shape[0].distance.max() == 1
    assert pdf.shape[1] == len(pdf.columns) == len(df.columns)

    # The `size` member should be a sensitive value
    assert isinstance(pdf.size, pj.Prisoner)
    assert pdf.size._value == df.size
    assert pdf.size.distance.max() == len(pdf.columns) == len(df.columns)

    # Builtin `len()` function should raise an error because it must be an integer value
    with pytest.raises(pj.DPError):
        len(pdf)

def test_priv_dataframe_comp() -> None:
    pdf, df = load_dataframe()

    # A non-sensitive value should be successfully compared against a private dataframe
    assert isinstance(pdf == 3, ppd.PrivDataFrame)
    assert isinstance(pdf != 3, ppd.PrivDataFrame)
    assert isinstance(pdf <  3, ppd.PrivDataFrame)
    assert isinstance(pdf <= 3, ppd.PrivDataFrame)
    assert isinstance(pdf >  3, ppd.PrivDataFrame)
    assert isinstance(pdf >= 3, ppd.PrivDataFrame)
    assert ((pdf == 3)._value == (df == 3)).all().all()
    assert ((pdf != 3)._value == (df != 3)).all().all()
    assert ((pdf <  3)._value == (df <  3)).all().all()
    assert ((pdf <= 3)._value == (df <= 3)).all().all()
    assert ((pdf >  3)._value == (df >  3)).all().all()
    assert ((pdf >= 3)._value == (df >= 3)).all().all()

    # A non-sensitive value should be successfully compared against a private series
    assert isinstance(pdf["a"] == 3, ppd.PrivSeries)
    assert isinstance(pdf["a"] != 3, ppd.PrivSeries)
    assert isinstance(pdf["a"] <  3, ppd.PrivSeries)
    assert isinstance(pdf["a"] <= 3, ppd.PrivSeries)
    assert isinstance(pdf["a"] >  3, ppd.PrivSeries)
    assert isinstance(pdf["a"] >= 3, ppd.PrivSeries)
    assert ((pdf["a"] == 3)._value == (df["a"] == 3)).all()
    assert ((pdf["a"] != 3)._value == (df["a"] != 3)).all()
    assert ((pdf["a"] <  3)._value == (df["a"] <  3)).all()
    assert ((pdf["a"] <= 3)._value == (df["a"] <= 3)).all()
    assert ((pdf["a"] >  3)._value == (df["a"] >  3)).all()
    assert ((pdf["a"] >= 3)._value == (df["a"] >= 3)).all()

    # An irrelevant, non-private dataframe should not be compared against a private dataframe
    with pytest.raises(TypeError): pdf == df
    with pytest.raises(TypeError): pdf != df
    with pytest.raises(TypeError): pdf <  df
    with pytest.raises(TypeError): pdf <= df
    with pytest.raises(TypeError): pdf >  df
    with pytest.raises(TypeError): pdf >= df

    # An irrelevant, non-private 2d array should not be compared against a private dataframe
    with pytest.raises(TypeError): pdf == [[list(range(len(pdf.columns)))] for x in range(5)]
    with pytest.raises(TypeError): pdf != [[list(range(len(pdf.columns)))] for x in range(5)]
    with pytest.raises(TypeError): pdf <  [[list(range(len(pdf.columns)))] for x in range(5)]
    with pytest.raises(TypeError): pdf <= [[list(range(len(pdf.columns)))] for x in range(5)]
    with pytest.raises(TypeError): pdf >  [[list(range(len(pdf.columns)))] for x in range(5)]
    with pytest.raises(TypeError): pdf >= [[list(range(len(pdf.columns)))] for x in range(5)]

    # An irrelevant, non-private array should not be compared against a private series
    with pytest.raises(TypeError): pdf["a"] == [0, 1, 2, 3, 4]
    with pytest.raises(TypeError): pdf["a"] != [0, 1, 2, 3, 4]
    with pytest.raises(TypeError): pdf["a"] <  [0, 1, 2, 3, 4]
    with pytest.raises(TypeError): pdf["a"] <= [0, 1, 2, 3, 4]
    with pytest.raises(TypeError): pdf["a"] >  [0, 1, 2, 3, 4]
    with pytest.raises(TypeError): pdf["a"] >= [0, 1, 2, 3, 4]

    x = pj.Prisoner(value=0, distance=pj.Distance(1), root_name=str(uuid.uuid4()))

    # A sensitive value should not be compared against a private dataframe
    with pytest.raises(TypeError): pdf == x
    with pytest.raises(TypeError): pdf != x
    with pytest.raises(TypeError): pdf <  x
    with pytest.raises(TypeError): pdf <= x
    with pytest.raises(TypeError): pdf >  x
    with pytest.raises(TypeError): pdf >= x

    # A sensitive value should not be compared against a private series
    with pytest.raises(TypeError): pdf["a"] == x
    with pytest.raises(TypeError): pdf["a"] != x
    with pytest.raises(TypeError): pdf["a"] <  x
    with pytest.raises(TypeError): pdf["a"] <= x
    with pytest.raises(TypeError): pdf["a"] >  x
    with pytest.raises(TypeError): pdf["a"] >= x

    # Sensitive dataframes of potentially different size should not be compared
    pdf_ = pdf[pdf["a"] >= 0]
    with pytest.raises(pj.DPError): pdf == pdf_
    with pytest.raises(pj.DPError): pdf != pdf_
    with pytest.raises(pj.DPError): pdf <  pdf_
    with pytest.raises(pj.DPError): pdf <= pdf_
    with pytest.raises(pj.DPError): pdf >  pdf_
    with pytest.raises(pj.DPError): pdf >= pdf_

    # Sensitive series of potentially different size should not be compared
    with pytest.raises(pj.DPError): pdf["a"] == pdf_["a"]
    with pytest.raises(pj.DPError): pdf["a"] != pdf_["a"]
    with pytest.raises(pj.DPError): pdf["a"] <  pdf_["a"]
    with pytest.raises(pj.DPError): pdf["a"] <= pdf_["a"]
    with pytest.raises(pj.DPError): pdf["a"] >  pdf_["a"]
    with pytest.raises(pj.DPError): pdf["a"] >= pdf_["a"]

def test_priv_dataframe_getitem() -> None:
    pdf, df = load_dataframe()

    # A single-column view should be successfully retrieved from a private dataframe
    assert isinstance(pdf["a"], ppd.PrivSeries)
    assert (pdf["a"]._value == df["a"]).all()

    # A multi-column view should be successfully retrieved from a private dataframe
    assert isinstance(pdf[["a", "b"]], ppd.PrivDataFrame)
    assert (pdf[["a", "b"]]._value == df[["a", "b"]]).all().all()

    # An irrelevant, non-sensitve bool vector should not be accepted for filtering (dataframe)
    with pytest.raises(TypeError):
        pdf[[True, True, False, False, True]]

    # An irrelevant, non-sensitve bool vector should not be accepted for filtering (series)
    with pytest.raises(TypeError):
        pdf["a"][[True, True, False, False, True]]

    # A bool-filtered view should be successfully retrieved from a private dataframe
    assert isinstance(pdf[pdf["a"] > 3], ppd.PrivDataFrame)
    assert (pdf[pdf["a"] > 3]._value == df[df["a"] > 3]).all().all()

    # A bool-filtered view should be successfully retrieved from a private series
    assert isinstance(pdf["a"][pdf["a"] > 3], ppd.PrivSeries)
    assert (pdf["a"][pdf["a"] > 3]._value == df["a"][df["a"] > 3]).all()

    # A sensitve bool vector of potentially different size should not be accepted for filtering (dataframe)
    pdf_ = pdf[pdf["a"] >= 0]
    with pytest.raises(pj.DPError):
        pdf[pdf_["a"] > 3]

    # A sensitve bool vector of potentially different size should not be accepted for filtering (series)
    with pytest.raises(pj.DPError):
        pdf["a"][pdf_["a"] > 3]

    x = pj.Prisoner(value=0, distance=pj.Distance(1), root_name=str(uuid.uuid4()))

    # A sensitive value should not be used as a column name
    with pytest.raises(TypeError):
        pdf[x]

    # A slice should not be used for selecting rows of a dataframe
    # TODO: it might be legal to accept a slice
    with pytest.raises(TypeError):
        pdf[2:5]

    # A slice should not be used for selecting rows of a series
    # TODO: it might be legal to accept a slice
    with pytest.raises(TypeError):
        pdf["a"][2:5]

def test_priv_dataframe_setitem() -> None:
    pdf, df = load_dataframe()

    # A non-sensitive value should be successfully assigned to a single-column view
    pdf["c"] = 10
    df["c"] = 10
    assert (pdf.columns == df.columns).all()
    assert (pdf._value == df).all().all()

    # A private series should be successfully assigned to a single-column view
    pdf["d"] = pdf["a"]
    df["d"] = df["a"]
    assert (pdf.columns == df.columns).all()
    assert (pdf._value == df).all().all()

    # A non-sensitive value should be successfully assigned to a multi-column view
    pdf[["e", "f"]] = 100
    df[["e", "f"]] = 100
    assert (pdf.columns == df.columns).all()
    assert (pdf._value == df).all().all()

    # A private dataframe should be successfully assigned to a multi-column view
    pdf[["g", "h"]] = pdf[["a", "b"]]
    df[["g", "h"]] = df[["a", "b"]]
    assert (pdf.columns == df.columns).all()
    assert (pdf._value == df).all().all()

    # A non-sensitive value should be successfully assigned to a bool-filtered view
    pdf[pdf["a"] < 3] = 8
    df[df["a"] < 3] = 8
    assert (pdf.columns == df.columns).all()
    assert (pdf._value == df).all().all()

    # A non-sensitive array should be successfully assigned to a bool-filtered view
    pdf[pdf["a"] == 8] = list(range(len(pdf.columns)))
    df[df["a"] == 8] = list(range(len(df.columns)))
    assert (pdf.columns == df.columns).all()
    assert (pdf._value == df).all().all()

    # A private dataframe should be successfully assigned to a bool-filtered view
    pdf[pdf["a"] < 3] = pdf[pdf["a"] < 4]
    df[df["a"] < 3] = df[df["a"] < 4]
    assert (pdf.columns == df.columns).all()
    assert (pdf._value == df).all().all()

    x = pj.Prisoner(value=0, distance=pj.Distance(1), root_name=str(uuid.uuid4()))

    # A sensitive value should not be assigned to a single-column view
    with pytest.raises(TypeError):
        pdf["x"] = x

    # A sensitive value should not be assigned to a multi-column view
    with pytest.raises(TypeError):
        pdf[["x", "y"]] = x

    # A sensitive value should not be assigned to a bool-filtered view
    with pytest.raises(TypeError):
        pdf[pdf["a"] > 3] = x

    # An irrelevant, non-private array should not be assigned to a column view
    with pytest.raises(TypeError):
        pdf["x"] = [1, 2, 3, 4, 5]

    # An irrelevant, non-private 2d array should not be assigned to a bool-filtered view
    with pytest.raises(TypeError):
        pdf[pdf["a"] > 3] = [[list(range(len(pdf.columns)))] for x in range(5)]

    # A sensitive value should not be used as a column name
    with pytest.raises(TypeError):
        pdf[x] = 10

    # A sensitve bool vector of potentially different size should not be accepted for filtering (dataframe)
    pdf_ = pdf[pdf["a"] >= 0]
    with pytest.raises(pj.DPError):
        pdf[pdf_["a"] > 3] = 10

    # A sensitve bool vector of potentially different size should not be accepted for filtering (series)
    with pytest.raises(pj.DPError):
        pdf["a"][pdf_["a"] > 3] = 10

    # A slice should not be used for selecting rows
    # TODO: it might be legal to accept a slice
    with pytest.raises(TypeError):
        pdf[2:5] = 0

def test_priv_dataframe_replace() -> None:
    pdf, df = load_dataframe()

    # Default behaviour
    assert (pdf.replace(4, 10)._value == df.replace(4, 10)).all().all()

    # Special behaviour when `inplace=True`
    assert pdf.replace(3, 10, inplace=True) == None
    assert (pdf._value == df.replace(3, 10)).all().all()

def test_priv_series_replace() -> None:
    pdf, df = load_dataframe()

    # Default behaviour
    assert (pdf["b"].replace(4, 10)._value == df["b"].replace(4, 10)).all()

    # Special behaviour when `inplace=True`
    assert pdf["b"].replace(3, 10, inplace=True) == None
    assert (pdf["b"]._value == df["b"].replace(3, 10)).all()

def test_priv_dataframe_dropna() -> None:
    pdf, df = load_dataframe()

    pdf.replace(3, np.nan, inplace=True)
    df.replace(3, np.nan, inplace=True)

    # Default behaviour
    assert (pdf.dropna()._value == df.dropna()).all().all()

    # Should return an error with ignore_index=True
    with pytest.raises(pj.DPError):
        pdf.dropna(ignore_index=True)

    # Special behaviour when `inplace=True`
    assert pdf.dropna(inplace=True) == None
    assert (pdf._value == df.dropna()).all().all()

def test_priv_series_dropna() -> None:
    pdf, df = load_dataframe()

    pdf.replace(3, np.nan, inplace=True)
    df.replace(3, np.nan, inplace=True)

    # Default behaviour
    assert (pdf["b"].dropna()._value == df["b"].dropna()).all()

    # Should return an error with ignore_index=True
    with pytest.raises(pj.DPError):
        pdf["b"].dropna(ignore_index=True)

    # TODO: this fails because of the original pandas bug?
    # # Special behaviour when `inplace=True`
    # assert pdf["b"].dropna(inplace=True) == None
    # assert (pdf["b"]._value == df["b"].dropna()).all()

def test_priv_series_value_counts() -> None:
    pdf, df = load_dataframe()

    # Should return an error without arguments
    with pytest.raises(pj.DPError):
        pdf["b"].value_counts()

    # Should return an error with `sort=True`
    with pytest.raises(pj.DPError):
        pdf["b"].value_counts(values=[1, 2, 3, 4, 5])

    # Should return an error without specifying values
    with pytest.raises(pj.DPError):
        pdf["b"].value_counts(sort=False)

    # Should return correct counts when all possible values are provided
    values = [2, 3, 4]
    counts = pdf["b"].value_counts(sort=False, values=values)
    assert_equal_sensitive_series(counts, pd.Series({2: 1, 3: 1, 4: 3}))
    assert counts.iloc[0].distance.max() == 1

    # Should return correct counts when only a part of possible values are provided
    values = [3, 4]
    counts = pdf["b"].value_counts(sort=False, values=values)
    assert_equal_sensitive_series(counts, pd.Series({3: 1, 4: 3}))
    assert counts.iloc[0].distance.max() == 1

    # Should return correct counts when non-existent values are provided
    values = [1, 3, 4, 5]
    counts = pdf["b"].value_counts(sort=False, values=values)
    assert_equal_sensitive_series(counts, pd.Series({1: 0, 3: 1, 4: 3, 5: 0}))
    assert counts.iloc[0].distance.max() == 1

    # Should be able to get a sensitive value from a sensitive series
    c4 = counts[4]
    assert isinstance(c4, pj.Prisoner)
    assert c4.distance.max() == 1
    assert c4._value == 3

    # Should be able to get a sensitive view from a sensitive series
    c3 = counts[1:3][3]
    c4 = counts[1:3][4]
    assert isinstance(c3, pj.Prisoner)
    assert isinstance(c4, pj.Prisoner)
    assert c3.distance.max() == c4.distance.max() == 1
    assert c3._value == 1
    assert c4._value == 3

def test_crosstab() -> None:
    pdf, df = load_dataframe()

    rowvalues = [1, 2, 3, 4, 5]
    colvalues = [1, 2, 3, 4, 5]

    # Should raise an error without rowvalues/colvalues
    with pytest.raises(pj.DPError): ppd.crosstab(pdf["a"], pdf["b"])
    with pytest.raises(pj.DPError): ppd.crosstab(pdf["a"], pdf["b"], rowvalues=rowvalues)
    with pytest.raises(pj.DPError): ppd.crosstab(pdf["a"], pdf["b"], colvalues=colvalues)

    # Should raise an error with margins=True
    with pytest.raises(pj.DPError):
        ppd.crosstab(pdf["a"], pdf["b"], rowvalues=rowvalues, colvalues=colvalues, margins=True)

    # Should raise an error with series of potentially different size
    pdf_ = pdf[pdf["a"] >= 0]
    with pytest.raises(pj.DPError):
        ppd.crosstab(pdf["a"], pdf_["b"], rowvalues=rowvalues, colvalues=colvalues)

    # Should return correct counts when all possible values are provided
    counts = ppd.crosstab(pdf["a"], pdf["b"], rowvalues=rowvalues, colvalues=colvalues)
    ans = pd.DataFrame([[0, 1, 0, 0, 0],
                        [0, 0, 0, 1, 0],
                        [0, 0, 0, 1, 0],
                        [0, 0, 0, 1, 0],
                        [0, 0, 1, 0, 0]],
                       index=rowvalues, columns=colvalues)
    assert_equal_sensitive_dataframes(counts, ans)

    pj.laplace_mechanism(counts, eps=1.0)

def test_cut() -> None:
    pdf, df = load_dataframe()

    # Default behaviour
    assert isinstance(ppd.cut(pdf["a"], bins=[0, 3, 6]), ppd.PrivSeries)
    assert (ppd.cut(pdf["a"], bins=[0, 3, 6])._value == pd.cut(df["a"], bins=[0, 3, 6])).all()

    # Should raise an error with a scalar bins
    with pytest.raises(TypeError):
        ppd.cut(pdf["a"], bins=2)

def test_dataframe_groupby() -> None:
    pdf, df = load_dataframe()

    # Should raise an error without `keys`
    with pytest.raises(pj.DPError):
        pdf.groupby("b")

    # Should be able to group by a single column
    keys = [1, 2, 4]
    groups = pdf.groupby("b", keys=keys)
    assert len(groups) == len(keys)

    # Should be able to loop over groups
    for i, (key, pdf_) in enumerate(groups):
        assert key == keys[i]
        assert isinstance(pdf_, ppd.PrivDataFrame)

    # A group with key=1 should be empty but its columns and dtypes should match the original
    assert isinstance(groups.get_group(1), ppd.PrivDataFrame)
    assert len(groups.get_group(1)._value) == 0
    assert (groups.get_group(1).columns == pdf.columns).all()
    assert (groups.get_group(1).dtypes == pdf.dtypes).all()

    # Check for a group with key=2
    assert isinstance(groups.get_group(2), ppd.PrivDataFrame)
    assert (groups.get_group(2)._value == pd.DataFrame({"a": [1], "b": [2]}, index=[0])).all().all()
    assert (groups.get_group(2).columns == pdf.columns).all()
    assert (groups.get_group(2).dtypes == pdf.dtypes).all()

    # Check for a group with key=4
    assert isinstance(groups.get_group(4), ppd.PrivDataFrame)
    assert (groups.get_group(4)._value == pd.DataFrame({"a": [2, 3, 4], "b": [4, 4, 4]}, index=[1, 2, 3])).all().all()
    assert (groups.get_group(4).columns == pdf.columns).all()
    assert (groups.get_group(4).dtypes == pdf.dtypes).all()

def test_privacy_budget() -> None:
    pdf, df = load_dataframe()

    eps = 0.1
    pdf1 = pdf[pdf["b"] >= 3]
    counts = pdf1["b"].value_counts(sort=False, values=[3, 4, 5])

    pj.laplace_mechanism(counts, eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps

    pj.laplace_mechanism(counts[4], eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 2

    pdf2 = pdf[pdf["a"] >= 3]

    pj.laplace_mechanism(pdf2.shape[0], eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 3

    # Privacy budget for different data sources should be managed independently
    pdf_, df_ = load_dataframe()

    pj.laplace_mechanism(pdf_.shape[0], eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 3
    assert pj.consumed_privacy_budget()[pdf_.root_name()] == eps

def test_privacy_budget_parallel_composition() -> None:
    pdf, df = load_dataframe()

    eps = 0.1

    # value_counts()
    counts = pdf["b"].value_counts(sort=False, values=[2, 3, 4])

    pj.laplace_mechanism(counts, eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps

    pj.laplace_mechanism(counts[2], eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 2

    pj.laplace_mechanism(counts[3], eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 2

    pj.laplace_mechanism(counts[3] + counts[4], eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 3

    pj.laplace_mechanism(counts[2] - counts[4], eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 3

    # crosstab()
    crosstab = ppd.crosstab(pdf["a"], pdf["b"], rowvalues=[1, 2, 3, 4, 5], colvalues=[1, 2, 3, 4, 5])

    for idx in crosstab.index:
        for col in crosstab.columns:
            pj.laplace_mechanism(crosstab.loc[idx, col], eps=eps)

    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 4

    s = pj.SensitiveInt(0)
    for idx in crosstab.index:
        for col in crosstab.columns:
            s += crosstab.loc[idx, col]
    assert s.distance.max() == 1
    pj.laplace_mechanism(s, eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 5

    assert crosstab.loc[1, 5].distance.max() == 1 # type: ignore
    pj.laplace_mechanism(crosstab.loc[1, 5], eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 6

    # groupby()
    s = pj.SensitiveInt(0)
    for key, pdf_ in pdf.groupby("b", keys=[1, 2, 3, 4, 5]):
        s += pdf_.shape[0]
    assert s.distance.max() == 1
    assert s._value == len(pdf._value)
    pj.laplace_mechanism(s, eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 7

    s = pj.SensitiveInt(0)
    for key, pdf_ in pdf.groupby("b", keys=[1, 2, 3, 4, 5]):
        s += key * pdf_.shape[0]
    assert s.distance.max() == 5
    pj.laplace_mechanism(s, eps=eps)
    assert pj.consumed_privacy_budget()[pdf.root_name()] == eps * 8
